pub struct GravityFieldConfig {
pub filepath: PathBuf,
pub gunzipped: bool,
pub degree: usize,
pub order: usize,
}Expand description
Configuration holder for gravity field.
Data is first loaded as a SHADR, if that fails, Nyx will try to load it as a COF file.
Fields
filepath: PathBufPath to the file, relative to the current working director
gunzipped: boolSet to true if the data is gunzipped
degree: usizeDesired degree
order: usizeDesired order